/**
 * Header file for hash_table (GPU hash table implementation).
 * Compressed version (32-bits).
 *
 * Loosely based on the work of Nathan Cassee (TU/e).
 *
 * Part of Master Project 'Porting tree-based hash table compression to GPGPU model checking', 
 * University of Twente (UT).
 *
 * @author  Danny Bergsma (s0113360)
 * @version 3.0
 */

#ifndef HASH_TABLE_H
#define HASH_TABLE_H


	// -- INCLUDES--


#include <stdint.h>


	// -- TYPEDEFS --


/** Type of elements and indices in hash table. */
typedef uint32_t inttype, indextype;


	// -- LITERALS/CONSTANTS --


/** Number of threads in a warp. */
static const unsigned int WARP_SIZE = 32;

/** Indicates restricted bit for root node. */
static const inttype ROOT_BIT_32    = (((inttype) 1) << ((sizeof(inttype) * 8) - 1));
/** Restricted value that indicates an empty entry in hash table. */
static const inttype EMPTY_ENTRY_32 = ~ROOT_BIT_32;

/** Maximum number of entries. */
static const indextype MAX_ENTRIES = (ROOT_BIT_32 - 1);

/** Maximum grid dimension (number of blocks). */
static const unsigned long MAX_GRID_DIM  = 2147483647ul;
/** Maximum block dimension (number of threads in a block). */
static const unsigned int  MAX_BLOCK_DIM = 1024;


	// -- FUNCTION DECLARATIONS --


/**
 * Initializes the (GPU) hash table.
 * @param    h_nEntries     number of (inttype) entries in hash table
 * @param    h_bucketSize   bucket size (number of hash table entries that form one bucket)
 * @param    h_vectorLength length of vector (number of elements)
 * @param    gridDim        grid dimension (number of blocks)
 * @param    blockDim       block dimension (number of threads in a block)
 * @requires (0 < h_nEntries <= MAX_ENTRIES) && ((h_nEntries % WARP_SIZE) == 0)
 * @requires (1 < h_bucketSize <= WARP_SIZE) && (h_bucketSize is power of 2)
 * @requires (0 < gridDim <= MAX_GRID_DIM) && (0 < blockDim <= MAX_BLOCK_DIM)
 */
void initialize(indextype h_nEntries, unsigned int h_bucketSize, unsigned int h_vectorLength,
                  unsigned long gridDim, unsigned int blockDim);

/**
 * Inserts the given (elements of) vectors into the (initialized) hash table, in parallel.
 * Some vectors may not be inserted, if the hash table is (considered) full. 
 * In that case a message will be printed.
 * @param    h_elements     array of the elements of the vectors to be inserted
 * @param    elementsLength total number of elements to be inserted
 * @param    gridDim        grid dimension (number of blocks)
 * @param    blockDim       block dimension (number of threads in a block)
 * @requires (elementsLength % vector length) == 0
 * @requires elementsLength <= length of h_elements array
 * @requires (\forall int elementIndex; (0 <= elementIndex < elementsLength);
 *                      ((h_elements[elementIndex] != EMPTY_ENTRY_32)
 *                         && !(h_elements[elementIndex] & ROOT_BIT_32)))
 * @requires (0 < gridDim <= MAX_GRID_DIM) && (0 < blockDim <= MAX_BLOCK_DIM)
 * @requires (blockDim % WARP_SIZE) == 0
 */
void insert(const inttype *h_elements, size_t elementsLength,
              unsigned long gridDim, unsigned int blockDim);

/**
 * Returns number of elements (non-empty entries) in (initialized) hash table, in parallel.
 * @param    gridDim  grid dimension (number of blocks)
 * @param    blockDim block dimension (number of threads in a block)
 * @return   number of elements in hash table
 * @requires (0 < gridDim <= MAX_GRID_DIM) && (0 < blockDim <= MAX_BLOCK_DIM)
 */
indextype elementCount(unsigned long gridDim, unsigned int blockDim);

/**
 * Returns number of vectors (root nodes) in (initialized) hash table, in parallel.
 * @param    gridDim  grid dimension (number of blocks)
 * @param    blockDim block dimension (number of threads in a block)
 * @return   number of vectors in hash table
 * @requires (0 < gridDim <= MAX_GRID_DIM) && (0 < blockDim <= MAX_BLOCK_DIM)
 */
indextype vectorCount(unsigned long gridDim, unsigned int blockDim);

/**
 * Saves dump of (initialized) hash table to dest.
 * @param    dest destination of dump
 * @param    size size of dump
 * @requires size <= size of dest
 */ 
void dump(inttype *dest, size_t size);

/**
 * Cleans up the (initialized) hash table. Hash table is then not initialized anymore.
 */
void cleanup();

#endif  // !HASH_TABLE_H