/**
 * Main program. Adds input file with state vectors to GPU hash table.
 * Compressed version.
 *
 * Loosely based on the work of Nathan Cassee (TU/e).
 *
 * Part of Master Project 'Porting tree-based hash table compression to GPGPU model checking',
 * University of Twente (UT).
 *
 * @author  Danny Bergsma (s0113360)
 * @version 2.1
 */


	// -- INCLUDES --


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hash_table.h"
#include "vector_file_reader.hpp"


	// -- LITERALS/CONSTANTS --


/** Default number of entries in hash table. */
static const indextype          DEFAULT_N_ENTRIES   = (((indextype) 1) << 26);  // 2^26
/** Default bucket size (number of hash table entries that form one bucket). */
static const unsigned int       DEFAULT_BUCKET_SIZE = 32;
/** Default grid dimension (number of blocks). */
static const unsigned long      DEFAULT_GRID_DIM    = 240;
/** Default block dimension (number of threads in a block). */
static const unsigned int       DEFAULT_BLOCK_DIM   = 256;

/** Flags for command line arguments. */
static const char FLAG_N_ENTRIES[]   = "-n",
                  FLAG_BUCKET_SIZE[] = "-s",
                  FLAG_GRID_DIM[]    = "-g",
                  FLAG_BLOCK_DIM[]   = "-b",
                  FLAG_DUMP_FILE[]   = "-d";

/** File mode for writing to dump file. */
static const char FILE_MODE[]       = "w";
/** Format of entry line in dump file. */
static const char DUMP_FILE_ENTRY[] = "0x%08x: 0x%08x (%d)\n";

/** Status messages. */
static const char STATUS_PARAMS[]                       = "TableSize: %zuB, BucketSize: %u, GridDim: %lu, BlockDim: %u\n",
                  STATUS_READING_INPUT[]                = "Reading input data...\n",
                  STATUS_VECTOR_INPUT[]                 = "VectorLength: %u, Vectors: %zu\n",
                  STATUS_INITIALIZING_TABLE[]           = "Initializing hash table...\n",
                  STATUS_RUNNING_INPUT_DATA[]           = "Running input data...\n",
                  STATUS_COLLECTING_RESULTS[]           = "Collecting results...\n",
                  STATUS_ELEMENTS_IN_TABLE[]            = "Elements in table: %u\n",
                  STATUS_VECTORS_IN_TABLE[]             = "Vectors in table: %u\n",
                  STATUS_SAVING_DUMP[]                  = "Saving dump of table...\n",
                  STATUS_CLEANING_UP[]                  = "Cleaning up...\n";
/** Error messages for incorrect command line arguments. */
static const char ERROR_INVALID_COMMAND_LINE_ARGUMENT[] = "Invalid command line argument: %s\n",
                  ERROR_INVALID_NUMBER_OF_ENTRIES[]     = "Invalid number of entries (number of entries 1..%u and multiple of %u)\n",
                  ERROR_INVALID_BUCKET_SIZE[]           = "Invalid bucket size (bucket size 2..%u and should be power of 2)\n",
                  ERROR_INVALID_GRID_DIMENSION[]        = "Invalid grid dimension (grid dimension 1..%lu)\n",
                  ERROR_INVALID_BLOCK_DIMENSION[]       = "Invalid block dimension (block dimension 1..%u and should be multiple of %u)\n";
/** Error messages. */
static const char ERROR_RESTRICTED_ELEMENT[]            = "Vector file contains restricted element(s)\n",
                  ERROR_WRITING_TO_DUMP_FILE[]          = "Error writing to dump file\n";


	// -- MAIN FUNCTION --


/**
 * Main function.
 * 1) Processes and checks command line arguments (see README).
 * 2) Reads input file and checks for validity (see README).
 * 3) Initializes GPU hash table.
 * 4) Runs input data (adding state vectors to GPU hash table).
 * 5) Collects and prints results (number of elements/vectors in GPU hash table).
 * 6) Saves dump of GPU hash table to file (if requested).
 * 7) Cleans up.
 * @param    argc number of command line arguments (including name of executable)
 * @param    argv command line arguments (array of strings)
 * @return   exit code (code >= 1 means error occurred)
 * @requires argc    >= 1
 * @ensures  \result >= 0
 */
int main(int argc, char **argv) {
	// setting defaults:
	char *inputFileName = NULL,        
	     *dumpFileName  = NULL;
	indextype nEntries   = DEFAULT_N_ENTRIES;
	int       bucketSize = DEFAULT_BUCKET_SIZE;
	long      gridDim    = DEFAULT_GRID_DIM;
	int       blockDim   = DEFAULT_BLOCK_DIM;
	
	// processing command line arguments:
	int argumentIndex = 1;  // ignoring first command line argument (name of executable)
	while (argumentIndex < argc) {
		if ((argumentIndex + 1) < argc) {  // at least two arguments left (flag + data)
			if      (strcmp(argv[argumentIndex], FLAG_N_ENTRIES)   == 0)
				nEntries     = atol(argv[++argumentIndex]);
			else if (strcmp(argv[argumentIndex], FLAG_BUCKET_SIZE) == 0)
				bucketSize   = atoi(argv[++argumentIndex]);
			else if (strcmp(argv[argumentIndex], FLAG_GRID_DIM)    == 0)
				gridDim      = atol(argv[++argumentIndex]);
			else if (strcmp(argv[argumentIndex], FLAG_BLOCK_DIM)   == 0)
				blockDim     = atoi(argv[++argumentIndex]);
			else if (strcmp(argv[argumentIndex], FLAG_DUMP_FILE)   == 0)
				dumpFileName =      argv[++argumentIndex];
			else
				fprintf(stderr, ERROR_INVALID_COMMAND_LINE_ARGUMENT, argv[argumentIndex]);
		} else  // last argument
			inputFileName = argv[argumentIndex];
		
		argumentIndex++;
	}
	// checking command line arguments:
	if ((nEntries   <= 0) || (MAX_ENTRIES   < nEntries)   || (nEntries % WARP_SIZE)) {
		fprintf(stderr, ERROR_INVALID_NUMBER_OF_ENTRIES, MAX_ENTRIES, WARP_SIZE);
		exit(1);
	}
	if ((bucketSize <= 1) || (WARP_SIZE     < bucketSize) || (bucketSize & (bucketSize - 1))) {
	  // (bucketSize & (bucketSize - 1)) <==> (bucketSize is not power of 2)
		fprintf(stderr, ERROR_INVALID_BUCKET_SIZE, WARP_SIZE);
		exit(1);
	}
	if ((gridDim    <= 0) || (MAX_GRID_DIM  < gridDim)) {
		fprintf(stderr, ERROR_INVALID_GRID_DIMENSION, MAX_GRID_DIM);
		exit(1);
	}
	if ((blockDim   <= 0) || (MAX_BLOCK_DIM < blockDim)   || (blockDim % WARP_SIZE)) {
	  // ((blockDim % WARP_SIZE) != 0) <==> (blockDim is not multiple of WARP_SIZE)
		fprintf(stderr, ERROR_INVALID_BLOCK_DIMENSION, MAX_BLOCK_DIM, WARP_SIZE);
		exit(1);
	}
	
	// printing parameters:
	size_t tableSize = (nEntries * sizeof(inttype));
	printf(STATUS_PARAMS, tableSize, bucketSize, gridDim, blockDim);
	
	// reading input file:
	printf(STATUS_READING_INPUT);
	Vectors input = readVectorFile(inputFileName);
	printf(STATUS_VECTOR_INPUT, input.vectorLength, input.nVectors);
	// checking input data (restricted elements):
	size_t elementsLength = (input.nVectors * input.vectorLength);
	for (size_t elementIndex = 0; (elementIndex < elementsLength); elementIndex++)
		if ((input.elements[elementIndex] == EMPTY_ENTRY_32)
			  || (input.elements[elementIndex] & ROOT_BIT_32)) {
			fprintf(stderr, ERROR_RESTRICTED_ELEMENT);
			exit(1);
		}
	
	// initializing hash table:
	printf(STATUS_INITIALIZING_TABLE);
	initialize(nEntries, bucketSize, input.vectorLength, gridDim, blockDim);
	
	// running input data:
	if (input.nVectors > 0) {
		printf(STATUS_RUNNING_INPUT_DATA);
		insert(input.elements, elementsLength, gridDim, blockDim);
		input.nVectors = 0;
		free(input.elements);
		input.elements = NULL;
	}
	
	// collecting and printing results:
	printf(STATUS_COLLECTING_RESULTS);
	printf(STATUS_ELEMENTS_IN_TABLE, elementCount(gridDim, blockDim));
	printf(STATUS_VECTORS_IN_TABLE,  vectorCount(gridDim, blockDim));
	
	// saving dump of table:
	if (dumpFileName != NULL) {
		printf(STATUS_SAVING_DUMP);
		FILE *dumpFile = fopen(dumpFileName, FILE_MODE);
		if (dumpFile == NULL)
			fprintf(stderr, ERROR_WRITING_TO_DUMP_FILE);
		else {
			inttype *table = (inttype*) malloc(tableSize);
			dump(table, tableSize);
			for (indextype entryIndex = 0; (entryIndex < nEntries); entryIndex++)
				fprintf(dumpFile, DUMP_FILE_ENTRY, entryIndex, table[entryIndex], table[entryIndex]);
			free(table);
			fclose(dumpFile);
		}
	}
	
	// cleaning up:
	printf(STATUS_CLEANING_UP);
	cleanup();
	
	return 0;
}