/**
 * GPU hash table implementation.
 * Compressed version (64-bits), with linear probing (8 entries) and fixed bucket size (2).
 * No recursion, using global memory, each vector is fully processed by a constant-sized vector group (neighboured pairs).
 *
 * Loosely based on the work of Nathan Cassee (TU/e).
 *
 * Part of Master Project 'Porting tree-based hash table compression to GPGPU model checking', 
 * University of Twente (UT).
 *
 * @author  Danny Bergsma (s0113360)
 * @version 3.0
 */


	// -- INCLUDES --


#include "hash_table.h"

#include <stdio.h>

#include <cuda_runtime.h>

#include "mt64.h"


	// --- DEVICE CODE ---
	
	
		// -- INCLUDES --
	
	
	#include <device_launch_parameters.h>
	
	
		// -- ENUMS --
	
	
	/** Enumeration of possible return values of findOrPut(). */
	enum FoundOrPut {NEW, SEEN, FULL};
	
	
		// -- LITERALS/CONSTANTS --
	
	
	/** 64-bit version of the empty entry. */
	static const uint64_t EMPTY_ENTRY_64    = ((((uint64_t) EMPTY_ENTRY_32) << (sizeof(EMPTY_ENTRY_32) * 8)) | EMPTY_ENTRY_32);
	/** Bitmask to unroot 64-bit entries (two 32-bit entries). */
	static const uint64_t UNROOT_BITMASK_64 = ~((uint64_t) ROOT_BIT_32);  // little-endian
	
	/** Number of hash functions (bounded double hashing). */
	static const unsigned int N_HASH_FUNCTIONS              = 32;
	/** Number of constants per hash function. */
	static const unsigned int N_CONSTANTS_PER_HASH_FUNCTION = 3;
	/** Total number of hash constants. */
	static const unsigned int N_TOTAL_HASH_CONSTANTS        = (N_HASH_FUNCTIONS * N_CONSTANTS_PER_HASH_FUNCTION);
	/** Size of hash in bits. */
	static const unsigned int HASH_BITS                     = (sizeof(uint32_t) * 8);
	
	/** Number of 64-bit entries in a linear probe. */
	static const unsigned int LINEAR_PROBE_LENGTH = (WARP_SIZE / 8);
	
	/** Fixed bucket size. */
	static const unsigned int BUCKET_SIZE = 2;
	
	/** Status message when table is (considered) full. */
	#define STATUS_TABLE_FULL  "Table is (considered) full\n"
	
	
		// -- CONSTANTS --
	
	
	/** Constants used in hash functions. */
	__constant__ uint64_t     d_hashConstants[N_TOTAL_HASH_CONSTANTS];
	/** Hash table: array of hash table entries. */
	__constant__ inttype      *d_table;
	/** 
	 * Number of hash table entries.
	 * @invariant length of d_table array == d_tableLength
	 * @invariant d_tableLength == (d_nBuckets * 2)
	 * @invariant (0 < d_tableLength <= MAX_ENTRIES) && ((d_tableLength % WARP_SIZE) == 0)
	 */
	__constant__ indextype    d_tableLength;
	/* Number of buckets. */
	__constant__ uint32_t     d_nBuckets;
	/** Length of vector (number of elements in vectors). */
	__constant__ unsigned int d_vectorLength;
	/**
	 * Indicates whether vector length is odd or not.
	 * @invariant d_oddVectorLength == (((d_vectorLength % 2) != 0) && (d_vectorLength > 1))
	 */
	__constant__ unsigned int d_oddVectorLength;
	/**
	 * Number of nodes in the lowest full row of compression tree.
	 * @invariant (d_vectorLength > 0)  ==> ((0 < d_lowestFullRowLength <= d_vectorLength)
	 *                                         && d_lowestFullRowLength is power of 2)
	 * @invariant (d_vectorLength == 0) ==> (d_lowestFullRowLength == 0)
	 */
	__constant__ unsigned int d_lowestFullRowLength;
	/**
	 * Number of nodes in the incomplete, lowest row of compression tree.
	 * @invariant d_incompleteRowLength < (d_vectorLength / MAX_CHILDREN_BINARY_TREE)
	 */
	__constant__ unsigned int d_incompleteRowLength;
	
	
		// -- MACROS --
	
	
	/** Counts all hash table entries for which condition holds. */
	#define COUNT_DEVICE(condition)  {                                                                                      \
		indextype localCount = 0;                                                                                           \
		for (indextype entryIndex = ((((indextype) blockIdx.x) * blockDim.x) + threadIdx.x); (entryIndex < d_tableLength);  \
			  (entryIndex += (blockDim.x * gridDim.x)))                                                                     \
			if (condition)                                                                                                  \
				localCount++;                                                                                               \
		                                                                                                                    \
		atomicAdd(count, localCount);                                                                                       \
	}
	
	/**
	 * Index of thread in warp.
	 * @invariant 0 <= LANE < WARP_SIZE
	 */
	#define LANE  (threadIdx.x % WARP_SIZE)
	
	/**
	 * Index of vector group (in warp) the current thread belongs to.
	 * @invariant 0 <= VECTOR_GROUP_INDEX(vectorGroupSize) < WARP_SIZE
	 */
	#define VECTOR_GROUP_INDEX(vectorGroupSize)  (LANE / vectorGroupSize)
	
	/**
	 * Index in vector group that corresponds to the current thread.
	 * @invariant 0 <= INDEX_IN_VECTOR_GROUP(vectorGroupSize) < vectorGroupSize
	 */
	#define INDEX_IN_VECTOR_GROUP(vectorGroupSize)  (LANE % vectorGroupSize)
	
	/** Sets bits lower..(lower+amount-1) of bitmask to 1, all other bits to 0 (index LSB = 0). */
	#define SET_BITS(bitmask, lower, amount)  bitmask = ((uint32_t) (((1ull << (lower + amount)) - 1) ^ ((1ull << lower) - 1)));
	
		
		// -- KERNEL: INITIALIZING TABLE --
	
	
	/**
	 * Initializes the hash table by setting all entries to the empty entry value, in parallel.
	 */
	__global__ void initTable() {
		for (indextype entryIndex = ((((indextype) blockIdx.x) * blockDim.x) + threadIdx.x); (entryIndex < d_tableLength);
			  (entryIndex += (blockDim.x * gridDim.x)))  // grid-stride loop
			d_table[entryIndex] = EMPTY_ENTRY_32;
	}
	
	
		// -- KERNEL: COUNTING ELEMENTS --
	
	
	/** 
	 * Counts all elements in hash table, in parallel.
	 * @param count will (eventually) be set to total number of elements in hash table
	 */
	__global__ void countElements(indextype *count) {
		COUNT_DEVICE(d_table[entryIndex] != EMPTY_ENTRY_32)
	}
	
	
		// -- KERNEL: COUNTING VECTORS --
	
	
	/**
	 * Counts all vectors in hash table, in parallel.
	 * @param count will (eventually) be set to total number of vectors in hash table
	 */
	__global__ void countVectors(indextype *count) {
		COUNT_DEVICE(d_table[entryIndex] & ROOT_BIT_32)
	}
	
	
		// -- HELPER FUNCTIONS KERNEL INSERTING VECTORS: FIND_OR_PUT --
	
	
	/**
	 * Returns hash of node, using the hash function with index hashFunctionIndex.
	 * Hash is valid as index of hash table.
	 * @param    hashFunctionIndex index of hash function to be used
	 * @param    node64            node to be hashed
	 * @return   hash of node
	 * @requires hashFunctionIndex < N_HASH_FUNCTIONS
	 * @ensures  0 <= \result < (d_tableLength / MAX_CHILDREN_BINARY_TREE)
	 */
	__device__ indextype hash(unsigned int hashFunctionIndex, uint64_t node64) {
		unsigned int hashConstantsOffset = (hashFunctionIndex * N_CONSTANTS_PER_HASH_FUNCTION);

		return ((((((d_hashConstants[hashConstantsOffset] + (node64 >> HASH_BITS)) * (d_hashConstants[hashConstantsOffset + 1] + ((indextype) node64)))
			+ d_hashConstants[hashConstantsOffset + 2]) >> HASH_BITS) * d_nBuckets) >> HASH_BITS);
	}
	
	/**
	 * Tries to add node to hash table, using bounded double hashing and linear probing.
	 * Returns index of node in hash table; EMPTY_ENTRY_32 is returned if table is full.
	 * @param    node node to be inserted into table
	 * @return   index of node in hash table; EMPTY_ENTRY_32 is returned if table is full
	 * @requires d_vectorLength > 0
	 * @requires length of node == min(d_vectorLength, MAX_CHILDREN_BINARY_TREE)
	 * @requires (\forall int elementIndex; 0 <= elementIndex < length of node;
	 *                      (node[elementIndex] != EMPTY_ENTRY_32)) && !(node[elementIndex] & ROOT_BIT_32)))
	 */
	__device__ indextype findOrPut(const inttype *node) {
		indextype result = EMPTY_ENTRY_32;  // assume full table (or considered full)
		
		uint64_t node64 = (((d_vectorLength == 1) ? 0 : (((uint64_t) node[1]) << HASH_BITS)) | node[0]);
		  // needed for alignment + padding
		
		unsigned int hashFunctionIndex = 0;
		while ((result == EMPTY_ENTRY_32) && (hashFunctionIndex < N_HASH_FUNCTIONS)) {
			indextype entryIndex = hash(hashFunctionIndex, node64);
			
			unsigned int probeIndex = 0;
			do {
				uint64_t *entryAddress = (uint64_t*) (d_table + (MAX_CHILDREN_BINARY_TREE * entryIndex)),
				         entry         =             (*entryAddress & UNROOT_BITMASK_64);  // unset root bit
				
				if (entry == node64)
					result = entryIndex;  // + exit both loops
				else {  // vector was not found ==> try to insert vector
					entry = (atomicCAS((unsigned long long*) entryAddress, EMPTY_ENTRY_64, node64)
					           & UNROOT_BITMASK_64);  // unset root bit
					if ((entry == EMPTY_ENTRY_64) || (entry == node64))
						result = entryIndex;  // + exit both loops
				}
				
				entryIndex = (((entryIndex / LINEAR_PROBE_LENGTH) * LINEAR_PROBE_LENGTH) 
				                + ((entryIndex + 1) % LINEAR_PROBE_LENGTH));
				probeIndex++;
			} while ((result == EMPTY_ENTRY_32) && (probeIndex < LINEAR_PROBE_LENGTH));
			
			hashFunctionIndex++;  // try next hash function (double hashing)
		}
		
		return result;
	}
	
	/**
	 * Tries to add vector to hash table, using tree-based hash table compression. Returns index of vector in hash table.
	 * If table is (considered) full, EMPTY_ENTRY_32 is returned.
	 * @param    vector          vector to be inserted into table
	 * @param    buildTree       location in memory where the tree can be built-up
	 * @param    vectorGroupMask bitmask of vector group the current thread belongs to
	 * @return   index of vector in hash table
	 * @requires d_vectorLength > 0
	 * @requires length of vector == d_vectorLength
	 * @requires (\forall int elementIndex; 0 <= elementIndex < d_vectorLength;
	 *                      (vector[elementIndex] != EMPTY_ENTRY_32)) && !(vector[elementIndex] & ROOT_BIT_32)))
	 * @requires length of buildTree == d_lowestFullRowLength
	 */
	__device__ indextype treeNoRec(const inttype *vector, indextype *buildTree, uint32_t vectorGroupMask) {
		// processing incomplete, lowest row (if there is one) and lowest full row (i.e., base cases):
		unsigned int indexInVectorGroup = INDEX_IN_VECTOR_GROUP(d_lowestFullRowLength),
		             incompleteRowNodeIndex;
		if ((incompleteRowNodeIndex = (indexInVectorGroup * MAX_CHILDREN_BINARY_TREE)) < d_incompleteRowLength) {
			const inttype *subVector = (vector + (incompleteRowNodeIndex * MAX_CHILDREN_BINARY_TREE));

			int oddNode = (((incompleteRowNodeIndex + 1) == d_incompleteRowLength) && d_oddVectorLength);
			indextype node[] = {findOrPut(subVector), (oddNode ? subVector[MAX_CHILDREN_BINARY_TREE]
			                                                   : findOrPut(subVector + MAX_CHILDREN_BINARY_TREE))};
			buildTree[indexInVectorGroup] = (((node[0] != EMPTY_ENTRY_32) && (node[1] != EMPTY_ENTRY_32)) ? findOrPut(node) : EMPTY_ENTRY_32);
		} else {
			const inttype *subVector = (vector + d_incompleteRowLength
			                                   + (indexInVectorGroup * MAX_CHILDREN_BINARY_TREE));
			buildTree[indexInVectorGroup] = findOrPut(subVector);
		}
		
		// processing other rows bottom-up, in-place (neighboured pairs):
		unsigned int rowLength = (d_lowestFullRowLength / MAX_CHILDREN_BINARY_TREE),
		             stride    = 1;
		while (__all_sync(vectorGroupMask, (buildTree[indexInVectorGroup] != EMPTY_ENTRY_32)) && (rowLength >= 1)) {  // + tree row sync
			if (indexInVectorGroup < rowLength) {
				unsigned int firstChildIndex = (indexInVectorGroup * (stride * MAX_CHILDREN_BINARY_TREE));
				indextype node[] = {buildTree[firstChildIndex], buildTree[firstChildIndex + stride]};
				buildTree[firstChildIndex] = findOrPut(node);
			}
			
			rowLength /= MAX_CHILDREN_BINARY_TREE;
			stride    *= MAX_CHILDREN_BINARY_TREE;
		}
		
		// returning root node:
		return ((rowLength == 0) ? buildTree[0] : EMPTY_ENTRY_32);
	}
	
	/** 
	 * Tries to add vector to hash table, using tree-based hash table compression. 
	 * If it was already present, SEEN is returned. If it was not present, the vector is added and NEW is returned.
	 * If table is (considered) full, FULL is returned.
	 * @param    vector          vector to be inserted into table
	 * @param    buildTree       location in memory where the tree can be built-up
	 * @param    vectorGroupMask bitmask of vector group the current thread belongs to
	 * @return   result: SEEN (already present), NEW (was not present) or FULL (table is full)
	 * @requires d_vectorLength > 0
	 * @requires length of vector == d_vectorLength
	 * @requires (\forall int elementIndex; 0 <= elementIndex < d_vectorLength; 
	 *                      (vector[elementIndex] != EMPTY_ENTRY_32)) && !(vector[elementIndex] & ROOT_BIT_32)))
	 * @requires length of buildTree == d_lowestFullRowLength 
	 */
	__device__ FoundOrPut treeFindOrPut(const inttype *vector, indextype *buildTree, uint32_t vectorGroupMask) {
		FoundOrPut result = FULL;  // assume full table
		
		indextype rootIndex = treeNoRec(vector, buildTree, vectorGroupMask);
		if ((INDEX_IN_VECTOR_GROUP(d_lowestFullRowLength) == 0) && (rootIndex != EMPTY_ENTRY_32))
			result = ((atomicOr((d_table + (rootIndex * MAX_CHILDREN_BINARY_TREE)), ROOT_BIT_32) & ROOT_BIT_32) ? SEEN : NEW);
			  // check and set ROOT_BIT
		result = (FoundOrPut) __shfl_sync(vectorGroupMask, result, 0, d_lowestFullRowLength);
		
		return result;
	}
	
	
		// -- KERNEL: INSERTING VECTORS --
	
	
	/**
	 * Inserts the given (elements of) vectors into the (initialized) hash table, in parallel.
	 * Some vectors may not be inserted, if the hash table is (considered) full.
	 * In that case a message will be printed.
	 * @param    d_elements     array of (the elements of) the vectors to be inserted
	 * @param    elementsLength total number of elements to be inserted
	 * @param    d_buildTrees   location in global memory where the trees can be built-up
	 * @requires (elementsLength % vector length) == 0
	 * @requires elementsLength <= length of d_elements array
	 * @requires (\forall int elementIndex; 0 <= elementIndex < elementsLength; 
	 *                      (d_elements[elementIndex] != EMPTY_ENTRY_32))
	 *                         && !(d_elements[elementIndex] & ROOT_BIT_32)))
	 * @requires length of d_buildTrees == (gridDim.x * blockDim.x)
	 */
	__global__ void insertTable(const inttype *d_elements, size_t elementsLength, indextype *d_buildTrees) {
		uint32_t vectorGroupMask;
		/* setting bitmask to vector group of current thread: only the bits of those threads are set
		   that are part of the vector group of the current thread */
		SET_BITS(vectorGroupMask, (VECTOR_GROUP_INDEX(d_lowestFullRowLength) * d_lowestFullRowLength), d_lowestFullRowLength)
		
		// processing (elements of) vectors (by grid-stride loop):
		size_t vectorIndex  = (((((size_t) blockIdx.x) * blockDim.x) + threadIdx.x)           / d_lowestFullRowLength),
		       stride       = (((((size_t) gridDim.x)  * blockDim.x) / d_lowestFullRowLength) * d_vectorLength),
		       elementIndex = (vectorIndex * d_vectorLength);
		indextype *buildTree = (d_buildTrees + (vectorIndex * d_lowestFullRowLength));
		while (elementIndex < elementsLength) {
			if ((treeFindOrPut((d_elements + elementIndex), buildTree, vectorGroupMask) == FULL)
			  && (INDEX_IN_VECTOR_GROUP(d_lowestFullRowLength) == 0))
				printf(STATUS_TABLE_FULL);
			elementIndex += stride;
		}
	}
	
	
		// ---


	// -- LITERALS/CONSTANTS --


/** Error message for CUDA errors. */
static const char ERROR_CUDA[] = "Error %s at line %d in file %s\n";


	// -- VARIABLES --


/** Hash table: array of entries. */
inttype *table;


	// -- MACROS --


/** 
 * Helper/inline "function" that checks return values of CUDA calls.
 * Prints any errors and exits application if an error occurred.
 */
#define CUDA_CHECK_RETURN(value)  {                                                        \
	cudaError_t _m_cudaStat = value;                                                       \
	if (_m_cudaStat != cudaSuccess) {                                                      \
		fprintf(stderr, ERROR_CUDA, cudaGetErrorString(_m_cudaStat), __LINE__, __FILE__);  \
		exit(1);                                                                           \
	}                                                                                      \
}

/** Counts all hash table entries for which the associated condition of kernel holds. */
#define COUNT_HOST(kernel)  {                                      \
	indextype result, *count;                                      \
	CUDA_CHECK_RETURN(cudaMallocManaged(&count, sizeof(*count)));  \
	kernel<<<gridDim, blockDim>>>(count);                          \
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());                    \
	result = *count;                                               \
	CUDA_CHECK_RETURN(cudaFree(count));                            \
	return result;                                                 \
}


	// -- FUNCTIONS --


void initialize(indextype h_nEntries, unsigned int h_vectorLength,
                  unsigned long gridDim, unsigned int blockDim) {
	// allocating table in GPU memory and filling it with empty entries (on GPU):
	CUDA_CHECK_RETURN(cudaMalloc(&table, (h_nEntries * sizeof(*table))));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_table,       &table,      sizeof(table)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_tableLength, &h_nEntries, sizeof(h_nEntries)));
	initTable<<<gridDim, blockDim>>>();
	
	// generating hash constants and copying to GPU constant memory:
	uint64_t h_hashConstants[N_TOTAL_HASH_CONSTANTS];
	init_genrand64(time(NULL));
	for (unsigned int hashFunctionIndex = 0; (hashFunctionIndex < N_TOTAL_HASH_CONSTANTS); hashFunctionIndex++)
		h_hashConstants[hashFunctionIndex] = genrand64_int64();
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_hashConstants, h_hashConstants, sizeof(h_hashConstants)));
	
	// calculating number of nodes in lowest rows of compression tree and copying to GPU constant memory:
	unsigned int leastPowerOf2 = h_vectorLength;
	while (leastPowerOf2 & (leastPowerOf2 - 1))  // not power of 2
		leastPowerOf2--;
	unsigned int h_incompleteRowLength = (h_vectorLength - leastPowerOf2),
	             h_lowestFullRowLength = ((leastPowerOf2 + 1) / MAX_CHILDREN_BINARY_TREE);  // + 1 for vectorLength == 1
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_incompleteRowLength, &h_incompleteRowLength, sizeof(h_incompleteRowLength)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_lowestFullRowLength, &h_lowestFullRowLength, sizeof(h_lowestFullRowLength)));
	
	// copying (meta) table information to GPU constant memory:
	uint32_t     h_nBuckets        = (uint32_t) (h_nEntries     / BUCKET_SIZE);
	unsigned int h_oddVectorLength =            ((h_vectorLength % MAX_CHILDREN_BINARY_TREE) && (h_vectorLength > 1));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_nBuckets,        &h_nBuckets,        sizeof(h_nBuckets)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_vectorLength,    &h_vectorLength,    sizeof(h_vectorLength)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_oddVectorLength, &h_oddVectorLength, sizeof(h_oddVectorLength)));
	
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());  // wait till filling table with empty entries is completed
}

void insert(const inttype *h_elements, size_t elementsLength,
              unsigned long gridDim, unsigned int blockDim) {
	// copying input data to GPU:
	inttype *elements;
	size_t  elementsSize = (elementsLength * sizeof(*h_elements));
	CUDA_CHECK_RETURN(cudaMalloc(&elements, elementsSize));
	CUDA_CHECK_RETURN(cudaMemcpy(elements, h_elements, elementsSize, cudaMemcpyHostToDevice));
	
	// allocating memory for trees:
	indextype *h_buildTrees;
	CUDA_CHECK_RETURN(cudaMalloc(&h_buildTrees, ((((size_t) gridDim) * blockDim) * sizeof(*h_buildTrees))));
	
	// running input data from GPU:
	insertTable<<<gridDim, blockDim>>>(elements, elementsLength, h_buildTrees);
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());  // wait till insertion is completed
	
	// cleaning up input data/memory for trees:
	CUDA_CHECK_RETURN(cudaFree(elements));
	CUDA_CHECK_RETURN(cudaFree(h_buildTrees));
}

indextype elementCount(unsigned long gridDim, unsigned int blockDim) {
	COUNT_HOST(countElements)
}

indextype vectorCount(unsigned long gridDim, unsigned int blockDim) {
	COUNT_HOST(countVectors)
}

void dump(inttype *dest, size_t size) {
	CUDA_CHECK_RETURN(cudaMemcpy(dest, table, size, cudaMemcpyDeviceToHost));
}

void cleanup() {
	CUDA_CHECK_RETURN(cudaFree(table));
	table = NULL;
	CUDA_CHECK_RETURN(cudaDeviceReset());
}