/**
 * GPU hash table implementation.
 * Uncompressed (fixed) version.
 *
 * Stand-alone implementation of hash table implementation from GPUexplore (v1),
 * without replication and with Pair-multiply-shift (parallel) hash function,
 * with configurable bucket size.
 * Loosely based on the work of Nathan Cassee (TU/e).
 *
 * Part of Master Project 'Porting tree-based hash table compression to GPGPU model checking', 
 * University of Twente (UT).
 *
 * @author  Danny Bergsma (s0113360)
 * @version 3.0
 */


	// -- INCLUDES --


#include "hash_table.h"

#include <stdio.h>

#include <cuda_runtime.h>

#include "mt64.h"


	// --- DEVICE CODE ---
	
	
		// -- INCLUDES --
	
	
	#include <device_launch_parameters.h>
	
	
		// -- ENUMS --
	
	
	/** Enumeration of possible return values of findOrPut(). */
	enum FoundOrPut {NEW, SEEN, FULL};
	
	
		// -- LITERALS/CONSTANTS --
	
	
	/** Number of hash functions (bounded double hashing). */
	static const unsigned int N_HASH_FUNCTIONS              = 32;
	/** Number of constants per hash function. */
	static const unsigned int N_CONSTANTS_PER_HASH_FUNCTION = 33;
	/** Total number of hash constants. */
	static const unsigned int N_TOTAL_HASH_CONSTANTS        = (N_HASH_FUNCTIONS * N_CONSTANTS_PER_HASH_FUNCTION);
	/** Number of vector elements that are paired together for calculation of (a partial) hash. */
	static const unsigned int PAIR                          = 2;
	
	/** Number of bits required for storing writing bit plus padding (to ensure that it is different from EMPTY_ENTRY). */
	static const unsigned int N_WRITING_BITS = 2;
	
	/** Status message when table is (considered) full. */
	#define STATUS_TABLE_FULL  "Table is (considered) full\n"
	
	
		// -- CONSTANTS --
	
	
	/** Constants used in hash functions. */
	__constant__ uint64_t     d_hashConstants[N_TOTAL_HASH_CONSTANTS];
	/** Hash table: array of hash table entries. */
	__constant__ inttype      *d_table;
	/** 
	 * Number of hash table entries.
	 * @invariant length of d_table array == d_tableLength
	 * @invariant d_tableLength == (d_nBuckets * d_bucketSize)
	 * @invariant (d_tableLength > 0) && ((d_tableLength % WARP_SIZE) == 0)
	 */
	__constant__ indextype    d_tableLength;
	/** 
	 * Number of buckets.
	 * @invariant 0 < d_nBuckets <= MAX_BUCKETS
	 */
	__constant__ uint32_t     d_nBuckets;
	/**
	 * Number of hash table entries that form a bucket.
	 * @invariant (0 < d_bucketSize <= WARP_SIZE) && (d_bucketSize is power of 2)
	 */
	__constant__ unsigned int d_bucketSize;
	/**
	 * Length of vector (number of elements in vectors).
	 * @invariant d_vectorLength <= d_bucketSize
	 */
	__constant__ unsigned int d_vectorLength;
	/**
	 * Half of the length of vector (number of elements in vectors).
	 * @invariant d_halfVectorLength == (d_vectorLength / 2)
	 */
	__constant__ unsigned int d_halfVectorLength;
	/**
	 * Indicates whether vector length is odd or not.
	 * @invariant d_oddVectorLength == ((d_vectorLength % 2) != 0)
	 */
	__constant__ unsigned int d_oddVectorLength;
	
	
		// -- MACROS --
	
	
	/** 
	 * Index of thread in warp.
	 * @invariant 0 <= LANE < WARP_SIZE
	 */
	#define LANE  (threadIdx.x % WARP_SIZE)
	
	/** 
	 * Index of bucket group (in warp) the current thread belongs to.
	 * @invariant 0 <= BUCKET_GROUP_INDEX < WARP_SIZE
	 */
	#define BUCKET_GROUP_INDEX(bucketSize)  (LANE / bucketSize)
	
	/**
	 * Index (offset) of entry (in hash table bucket) that corresponds to the current thread.
	 * @invariant 0 <= ENTRY_INDEX_IN_BUCKET < bucketSize
	 */
	#define ENTRY_INDEX_IN_BUCKET(bucketSize)               (LANE % bucketSize)
	/**
	 * Index (not offset) of (complete/incomplete) vector slot (in hash table bucket) 
	 * that corresponds to the current thread.
	 * @invariant 0 <= SLOT_INDEX_IN_BUCKET < ceil(bucketSize / vectorLength)
	 */
	#define SLOT_INDEX_IN_BUCKET(bucketSize, vectorLength)  (ENTRY_INDEX_IN_BUCKET(bucketSize) / vectorLength)
	
	/**
	 * Index (offset) of element (in vector) that corresponds to the current thread.
	 * @invariant 0 <= ELEMENT_INDEX < vectorLength
	 */
	#define ELEMENT_INDEX(bucketSize, vectorLength)  (ENTRY_INDEX_IN_BUCKET(bucketSize) % vectorLength)
	
	/** Sets bits lower..(lower+amount-1) of bitmask to 1, all other bits to 0 (index LSB = 0). */
	#define SET_BITS(bitmask, lower, amount)  bitmask = ((uint32_t) (((1ull << (lower + amount)) - 1) ^ ((1ull << lower) - 1)));
	
	/** Returns true iff vector is found in slot of bucket group. */
	#define CHECK_SLOTS_FOR_VECTOR(vectorGroupMask, bucketGroupMask)                  \
		__any_sync(bucketGroupMask, __all_sync(vectorGroupMask, (entry == element)))
	/**
	 * Returns lane+1 of the first thread in the bucket group for which the predicate holds.
	 * Returns 0 if there are no such threads.
	 * @ensures 0 <= GET_FIRST_MATCHING_THREAD <= WARP_SIZE
	 */
	#define GET_FIRST_MATCHING_THREAD(bucketGroupMask, predicate)                     \
		__ffs(__ballot_sync(bucketGroupMask, predicate))
	
	
		// -- KERNEL: INITIALIZING TABLE --
	
	
	/**
	 * Initializes the hash table by setting all entries to the empty entry value, in parallel.
	 */
	__global__ void initTable() {
		for (indextype entryIndex = ((((indextype) blockIdx.x) * blockDim.x) + threadIdx.x);
		       (entryIndex < d_tableLength);
		       (entryIndex += (blockDim.x * gridDim.x)))  // grid-stride loop
			d_table[entryIndex] = EMPTY_ENTRY;
	}
	
	
		// -- KERNEL: COUNTING ELEMENTS --
	
	
	/** 
	 * Counts all elements in hash table, in parallel.
	 * @param count will (eventually) be set to total number of elements in hash table
	 */
	__global__ void countElements(indextype *count) {
		indextype localCount = 0;  // thread-local count
		for (indextype entryIndex = ((((indextype) blockIdx.x) * blockDim.x) + threadIdx.x);
		       (entryIndex < d_tableLength);
		       (entryIndex += (blockDim.x * gridDim.x)))  // grid-stride loop
			if (d_table[entryIndex] != EMPTY_ENTRY)
				localCount++;
		
		atomicAdd((unsigned long long*) count, localCount);  // add to global count
	}
	
	
		// -- HELPER FUNCTIONS KERNEL INSERTING VECTORS: FIND_OR_PUT --
	
	
	/**
	 * Returns hash of vector, using the hash function with index hashFunctionIndex.
	 * @param    hashFunctionIndex index of hash function to be used
	 * @param    vector            vector to be hashed
	 * @param    vectorGroupMask   bitmask of vector group the current thread belongs to
	 * @param    bucketGroupMask   bitmask of bucket group the current thread belongs to
	 * @return   hash of vector
	 * @requires hashFunctionIndex < N_HASH_FUNCTIONS
	 * @requires length of vector == d_vectorLength
	 */
	__device__ uint32_t hash(unsigned int hashFunctionIndex, const inttype *vector, uint32_t vectorGroupMask, uint32_t bucketGroupMask) {
	  // pair-multiply-shift (parallel)
		uint64_t hashTmp = 0;
		
		unsigned int hashConstantsOffset = (hashFunctionIndex * N_CONSTANTS_PER_HASH_FUNCTION);
		if (ENTRY_INDEX_IN_BUCKET(d_bucketSize) < d_vectorLength) {  // only first vector group should calculate hash
			// pair-multiply:
			if (ENTRY_INDEX_IN_BUCKET(d_bucketSize) < d_halfVectorLength)
				hashTmp = ((d_hashConstants[hashConstantsOffset + (ENTRY_INDEX_IN_BUCKET(d_bucketSize) * PAIR)]
				              + vector[(ENTRY_INDEX_IN_BUCKET(d_bucketSize) * PAIR) + 1])
				           * (d_hashConstants[hashConstantsOffset + ((ENTRY_INDEX_IN_BUCKET(d_bucketSize) * PAIR) + 1)]
				                + vector[ENTRY_INDEX_IN_BUCKET(d_bucketSize) * PAIR]));
			else if ((ENTRY_INDEX_IN_BUCKET(d_bucketSize) == d_halfVectorLength) && d_oddVectorLength)
			  // last + odd element: process separately
				hashTmp = (d_hashConstants[hashConstantsOffset + (ENTRY_INDEX_IN_BUCKET(d_bucketSize) * PAIR)]
				             * vector[(ENTRY_INDEX_IN_BUCKET(d_bucketSize) * PAIR)]);
			
			// parallel reduction:
			unsigned int rowLength = d_vectorLength;
			while (rowLength > 1) {
				rowLength /= PAIR;
				hashTmp += __shfl_xor_sync(vectorGroupMask, hashTmp, rowLength);
			}
		}
		
		if (ENTRY_INDEX_IN_BUCKET(d_bucketSize) == 0) {  // bucket group leader
			hashTmp += d_hashConstants[hashConstantsOffset + (N_CONSTANTS_PER_HASH_FUNCTION - 1)];
			hashTmp >>= HASH_BITS;  // shift
		}
		
		return ((uint32_t) __shfl_sync(bucketGroupMask, hashTmp, 0, d_bucketSize));  // get full hash from bucket group leader
	}
	
	/**
	 * Tries to add vector to hash table, using (partial) warp-the-line and bounded double hashing. 
	 * If it was already present, SEEN is returned. If it was not present, the vector is added and NEW is returned.
	 * If table is (considered) full, FULL is returned.
	 * @param    vector          vector to be inserted into table
	 * @param    vectorGroupMask bitmask of vector group the current thread belongs to
	 * @param    bucketGroupMask bitmask of bucket group the current thread belongs to
	 * @return   result: SEEN (already present), NEW (was not present) or FULL (table is full)
	 * @requires d_vectorLength > 0
	 * @requires length of vector == d_vectorLength
	 * @requires (\forall int elementIndex; 0 <= elementIndex < d_vectorLength; 
	 *                      ((vector[elementIndex] != EMPTY_ENTRY) && !(vector[elementIndex] & WRITING_BIT)))
	 */
	__device__ FoundOrPut findOrPut(const inttype *vector, uint32_t vectorGroupMask, uint32_t bucketGroupMask) {
		FoundOrPut result = FULL;  // assume full table (or considered full)
		
		inttype element = vector[ELEMENT_INDEX(d_bucketSize, d_vectorLength)];
		
		unsigned int hashFunctionIndex = 0;
		while ((result == FULL) && (hashFunctionIndex < N_HASH_FUNCTIONS)) {
			uint32_t  hashValue  = hash(hashFunctionIndex, vector, vectorGroupMask, bucketGroupMask);
			indextype entryIndex = ((((((uint64_t) hashValue) * d_nBuckets) >> HASH_BITS) * d_bucketSize)
			                          + ENTRY_INDEX_IN_BUCKET(d_bucketSize));
			inttype   entry      = d_table[entryIndex];  // non-volatile
			
			if (CHECK_SLOTS_FOR_VECTOR(vectorGroupMask, bucketGroupMask))
			  // vector group(s) found complete vector in (slots of bucket of) hash table
				result = SEEN;  // + exit loop
			else {  // vector was not found ==> try to insert vector into first empty slot in bucket (or check writing bucket)
				volatile inttype *entryAddress = (d_table + entryIndex);  // pointer arithmetic; we now need volatile
				
				inttype writingInProgressValue = (WRITING_BIT | (((inttype) hashValue) >> N_WRITING_BITS));
				
				// firstEmptySlot is lane+1 of leader of first vector group with empty/writing slot:
				unsigned int firstEmptyOrWritingSlot =
				  GET_FIRST_MATCHING_THREAD(bucketGroupMask,
				    __all_sync(vectorGroupMask, ((entry == EMPTY_ENTRY) || (entry == element) || (entry == writingInProgressValue))));
				
				while ((result == FULL) && firstEmptyOrWritingSlot) {
				  // continue till vector is inserted/found or there are no empty/writing slots anymore
					firstEmptyOrWritingSlot--;  // firstEmptySlot is in 1..32 <-> LANE is in 0..31
					
					// vector group leader tries to claim its (empty) entry, i.e., the slot:
					if ((LANE == firstEmptyOrWritingSlot) && (entry == EMPTY_ENTRY))
						entry = atomicCAS((d_table + entryIndex), EMPTY_ENTRY, writingInProgressValue);
					
					// distribute claim result to all threads in bucket group:
					entry = __shfl_sync(bucketGroupMask, entry, firstEmptyOrWritingSlot);
					if (entry == EMPTY_ENTRY) {
					  // vector group leader successfully claimed its entry, i.e., the slot
						// other threads in vector group now write their elements to their entries:
						if ((firstEmptyOrWritingSlot < LANE) && (LANE < (firstEmptyOrWritingSlot + d_vectorLength)))
							*entryAddress = element;  // volatile
						__syncwarp(bucketGroupMask);
						// vector group leader writes its element and thus indicates that writing has completed: 
						if (LANE == firstEmptyOrWritingSlot)  // vector group leader
							*entryAddress = element;  // volatile
						
						result = NEW;  // + exit both loops
					} else {
						if (entry == writingInProgressValue) {  // potentially concurrent invocation inserting same vector
							if (LANE == firstEmptyOrWritingSlot)  // vector group leader
								while (*entryAddress == writingInProgressValue);  // spin-lock; volatile
							__syncwarp(bucketGroupMask);  // wait for spin-lock to complete
						}
						
						entry = *entryAddress;  // volatile
						
						if (CHECK_SLOTS_FOR_VECTOR(vectorGroupMask, bucketGroupMask))
							result = SEEN;  // + exit both loops
						else
							firstEmptyOrWritingSlot = GET_FIRST_MATCHING_THREAD(bucketGroupMask,
							  ((ELEMENT_INDEX(d_bucketSize, d_vectorLength) == 0) && ((entry == EMPTY_ENTRY) || (entry == writingInProgressValue))));
					}
				}
			}
			
			hashFunctionIndex++;  // try next hash function (double hashing)
		}
		
		return result;
	}
	
	
		// -- KERNEL: INSERTING VECTORS --
	
	
	/**
	 * Inserts the given (elements of) vectors into the (initialized) hash table, in parallel.
	 * Some vectors may not be inserted, if the hash table is (considered) full.
	 * In that case a message will be printed.
	 * @param    d_elements     array of (the elements of) the vectors to be inserted
	 * @param    elementsLength total number of elements to be inserted
	 * @requires (elementsLength % vector length) == 0
	 * @requires elementsLength <= length of d_elements array
	 * @requires (\forall int elementIndex; 0 <= elementIndex < elementsLength; 
	 *                      ((d_elements[elementIndex] != EMPTY_ENTRY) && !(d_elements[elementIndex] & WRITING_BIT)))
	 */
	__global__ void insertTable(const inttype *d_elements, size_t elementsLength) {
		if (((SLOT_INDEX_IN_BUCKET(d_bucketSize, d_vectorLength) + 1) * d_vectorLength) <= d_bucketSize) {
		  // slot/vector group not crossing bucket boundary
			uint32_t bucketGroupMask, vectorGroupMask;
			/* setting bitmask to bucket group of current thread: only the bits of those threads are set
			   that are part of the bucket group of the current thread */
			SET_BITS(bucketGroupMask,
			         (BUCKET_GROUP_INDEX(d_bucketSize) * d_bucketSize),
			         ((d_bucketSize / d_vectorLength) * d_vectorLength))
			/* setting bitmask to vector group of current thread: only the bits of those threads are set
			   that are part of the vector group of the current thread */
			SET_BITS(vectorGroupMask, 
			         ((BUCKET_GROUP_INDEX(d_bucketSize) * d_bucketSize) 
			         + (SLOT_INDEX_IN_BUCKET(d_bucketSize, d_vectorLength) * d_vectorLength)),
			         d_vectorLength)
			
			// processing (elements of) vectors (by grid-stride loop):
			size_t vectorIndex  = (((((size_t) blockIdx.x) * blockDim.x) + threadIdx.x)  / d_bucketSize),
			       stride       = (((((size_t) gridDim.x)  * blockDim.x) / d_bucketSize) * d_vectorLength),
			       elementIndex = (vectorIndex * d_vectorLength);
			while (elementIndex < elementsLength) {
				if ((findOrPut((d_elements + elementIndex), vectorGroupMask, bucketGroupMask) == FULL) 
				      && (ENTRY_INDEX_IN_BUCKET(d_bucketSize) == 0))  // only bucket group leader should print message
					printf(STATUS_TABLE_FULL);
				elementIndex += stride;
			}
		}
	}
	
	
		// ---


	// -- LITERALS/CONSTANTS --


/** Error message for CUDA errors. */
static const char ERROR_CUDA[] = "Error %s at line %d in file %s\n";


	// -- VARIABLES --


/** Hash table: array of entries. */
inttype *table;


	// -- MACROS --


/** 
 * Helper/inline "function" that checks return values of CUDA calls.
 * Prints any errors and exits application if an error occurred.
 */
#define CUDA_CHECK_RETURN(value)  {                                                        \
	cudaError_t _m_cudaStat = value;                                                       \
	if (_m_cudaStat != cudaSuccess) {                                                      \
		fprintf(stderr, ERROR_CUDA, cudaGetErrorString(_m_cudaStat), __LINE__, __FILE__);  \
		exit(1);                                                                           \
	}                                                                                      \
}


	// -- FUNCTIONS --


void initialize(indextype h_nEntries, unsigned int h_bucketSize, unsigned int h_vectorLength,
                  unsigned long gridDim, unsigned int blockDim) {
	// allocating table in GPU memory and filling it with empty entries (on GPU):
	CUDA_CHECK_RETURN(cudaMalloc(&table, (h_nEntries * sizeof(*table))));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_table,       &table,      sizeof(table)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_tableLength, &h_nEntries, sizeof(h_nEntries)));
	initTable<<<gridDim, blockDim>>>();
	
	// generating hash constants and copy to GPU constant memory:
	uint64_t h_hashConstants[N_TOTAL_HASH_CONSTANTS];
	init_genrand64(time(NULL));
	for (unsigned int hashFunctionIndex = 0; (hashFunctionIndex < N_TOTAL_HASH_CONSTANTS); hashFunctionIndex++)
		h_hashConstants[hashFunctionIndex] = genrand64_int64();
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_hashConstants, h_hashConstants, sizeof(h_hashConstants)));
	
	// copying (meta) table information to GPU constant memory:
	uint32_t     h_nBuckets = (uint32_t) (h_nEntries / h_bucketSize);
	unsigned int h_halfVectorLength = (h_vectorLength / PAIR),
	             h_oddVectorLength  = (h_vectorLength % PAIR);
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_nBuckets,         &h_nBuckets,         sizeof(h_nBuckets)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_bucketSize,       &h_bucketSize,       sizeof(h_bucketSize)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_vectorLength,     &h_vectorLength,     sizeof(h_vectorLength)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_halfVectorLength, &h_halfVectorLength, sizeof(h_halfVectorLength)));
	CUDA_CHECK_RETURN(cudaMemcpyToSymbol(d_oddVectorLength,  &h_oddVectorLength,  sizeof(h_oddVectorLength)));
	
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());  // wait till filling table with empty entries is completed
}

void insert(const inttype *h_elements, size_t elementsLength, 
              unsigned long gridDim, unsigned int blockDim) {	
	// copying input data to GPU:
	inttype *elements;
	size_t  elementsSize = (elementsLength * sizeof(*h_elements));
	CUDA_CHECK_RETURN(cudaMalloc(&elements, elementsSize));
	CUDA_CHECK_RETURN(cudaMemcpy(elements, h_elements, elementsSize, cudaMemcpyHostToDevice));
	
	// running input data from GPU:
	insertTable<<<gridDim, blockDim>>>(elements, elementsLength);
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());  // wait till insertion is completed
	
	// cleaning up input data:
	CUDA_CHECK_RETURN(cudaFree(elements));
}

indextype elementCount(unsigned long gridDim, unsigned int blockDim) {
	// using Unified Memory/Unified Virtual Addressing:
	indextype result, *count;
	CUDA_CHECK_RETURN(cudaMallocManaged(&count, sizeof(*count)));
	countElements<<<gridDim, blockDim>>>(count);
	CUDA_CHECK_RETURN(cudaDeviceSynchronize());  // wait till counting is completed
	result = *count;
	CUDA_CHECK_RETURN(cudaFree(count));
	return result;
}

void dump(inttype *dest, size_t size) {
	CUDA_CHECK_RETURN(cudaMemcpy(dest, table, size, cudaMemcpyDeviceToHost));
}

void cleanup() {
	CUDA_CHECK_RETURN(cudaFree(table));
	table = NULL;
	CUDA_CHECK_RETURN(cudaDeviceReset());
}